"""

"""

from lumo.base_classes.attr import attr

dic = attr()


dic.b.d = "asd"
dic.a = "asd"

print(dic.items())

# print(dic["b"])
# print("g" in dic)
print(dic.hash())
# c5e72bde733e687ce89c9798f825e04c