"""

"""

from lumo.base_classes.trickitems import NoneItem

none = NoneItem()
assert none > 0
assert none >= 0
assert none < 0
assert none <= 0
assert none != 0
assert none == None
assert none is not None
assert none/2 == 0.5
assert none*2 == 2
assert none+1 == 1
assert none-1 == -1

