import os


def split_file(file, line_chunk=1000, fmt='{:02d}'):
    """
    split one text file
        <filename>.<suf>
    into
        <filename>_<fmt_idx>.<suf>
    Returns:

    """
    c = 0

    pass
