"""

"""

from lumo.frame.experiment import globs

globs["nddatasets"] = "E:/Download/dataset"

globs.time = 0

print()
# globs.update("local", "a", 1)
# globs.list_config(True, True)

# print(globs.glob_fn)
